/* poller - front-end for waiting for I/O functions
   Copyright (C) 2010  Marek Behun

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU Lesser General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */

#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "config.h"
#include "poller.h"

#ifdef HAVE_EPOLL_WAIT

struct poller
{
  int epollfd;
  int nfds;
};

struct poller * pollers = NULL;
int npollers = 0;

int
poller_create ()
{
  struct poller * tmp;

  tmp = realloc (pollers, sizeof (struct poller) * (npollers + 1));
  if (tmp == NULL)
    {
      errno = ENOMEM;
      return (-1);
    }

  pollers = tmp;
  tmp = pollers + npollers;
  tmp->epollfd = epoll_create (1024);

  if (tmp->epollfd < 0)
    {
      int err = errno;
      pollers = realloc (pollers, sizeof (struct poller) * npollers);
      errno = err;
      return (-1);
    }

  tmp->nfds = 0;
  npollers ++;
  return tmp->epollfd;
}

static inline int
poller_index (int inst)
{
  int i;

  for (i = 0; i < npollers; i++)
    if (pollers[i].epollfd == inst)
      return i;

  errno = ENOENT;
  return (-1);
}

int
poller_destroy (int inst)
{
  int index = poller_index (inst);

  if (index < 0)
    return (-1);

  close (pollers[index].epollfd);

  if (index < -- npollers)
    memcpy (pollers + index, pollers + npollers, sizeof (struct poller));

  pollers = realloc (pollers, sizeof (struct poller) * npollers);
  return 0;
}

int
poller_add (int inst, int fd, pollerev_t events, pollerdata_t data)
{
  struct epoll_event event;
  int index = poller_index (inst);

  if (index < 0)
    return (-1);

  event.events = events;
  event.data = data;

  if (!epoll_ctl (inst, EPOLL_CTL_ADD, fd, &event))
    {
      pollers[index].nfds ++;
      return 0;
    }

  return (-1);
}

int
poller_modify (int inst, int fd, pollerev_t events, pollerdata_t data)
{
  struct epoll_event event;
  int index = poller_index (inst);

  if (index < 0)
    return (-1);

  event.events = events;
  event.data = data;
  return epoll_ctl (inst, EPOLL_CTL_MOD, fd, &event);
}

int
poller_remove (int inst, int fd)
{
  int ret, index = poller_index (inst);

  if (index < 0)
    return (-1);

  /* There is a bug in kernel versions before 2.6.9 in the epoll_ctl
     call: EPOLL_CTL_DEL require the event pointer to be a non-NULL
     pointer, even though it is ignored. That forces us to pass a
     pointer valid pointer as the fourth argument.  */

  ret = epoll_ctl (inst, EPOLL_CTL_DEL, fd, (struct epoll_event *) &ret);
  if (!ret)
    -- pollers[index].nfds;
  return ret;
}

int
poller_poll (int inst, int timeout, pollercb_t callback)
{
  struct epoll_event evs[8];
  int i, res, index = poller_index (inst);

  if (index < 0)
    return (-1);

  res = epoll_wait (inst, evs, 8, timeout);
  if (res <= 0)
    return res;

  for (i = 0; i < res; i++)
    callback (evs[i].data, evs[i].events);

  return res;
}

#else /* not HAVE_EPOLL_WAIT */

struct poller
{
  int id;
  struct pollfd * fds;
  pollerdata_t * data;
  int nfds;
};

struct poller * pollers = NULL;
int npollers = 0, pollers_newid = 0;

int
poller_create ()
{
  struct poller * tmp;

  tmp = realloc (pollers, (sizeof (struct poller)) * (npollers + 1));

  if (tmp == NULL)
    {
      errno = ENOMEM;
      return (-1);
    }

  pollers = tmp;
  tmp = pollers + npollers;
  tmp->id = pollers_newid ++;
  tmp->fds = NULL;
  tmp->data = NULL;
  tmp->nfds = 0;

  npollers ++;
  return (tmp->id);
}

static inline int
poller_index (int inst)
{
  int i;

  for (i = 0; i < npollers; i++)
    if (pollers[i].id == inst)
      return i;

  errno = ENOENT;
  return (-1);
}

int
poller_destroy (int inst)
{
  int index = poller_index (inst);
  struct poller * tmp;

  if (index < 0)
    return (-1);

  tmp = pollers + index;

  if (tmp->fds != NULL)
    free (tmp->fds);
  if (tmp->data != NULL)
    free (tmp->data);

  if (index < -- npollers)
    memcpy (pollers + index, pollers + npollers, sizeof (struct poller));

  pollers = realloc (pollers, sizeof (struct poller) * npollers);

  if (!npollers)
    pollers_newid = 0;

  return 0;
}

int
poller_add (int inst, int fd, pollerev_t events, pollerdata_t data)
{
  struct pollfd * new_fds;
  pollerdata_t * new_data;
  struct poller * tmp;
  int index = poller_index (inst);

  if (index < 0)
    return (-1);

  tmp = pollers + index;

  if (fd < 0)
    {
      errno = EBADF;
      return (-1);
    }

  new_fds = realloc (tmp->fds, sizeof (struct pollfd) * ++ tmp->nfds);
  if (new_fds == NULL)
    {
      -- tmp->nfds;
      errno = ENOMEM;
      return (-1);
    }

  new_data = realloc (tmp->data, sizeof (pollerdata_t) * tmp->nfds);
  if (new_data == NULL)
    {
      -- tmp->nfds;
      tmp->fds = realloc (new_fds, sizeof (struct pollfd) * tmp->nfds);
      errno = ENOMEM;
      return (-1);
    }

  tmp->fds = new_fds;
  tmp->data = new_data;

  tmp->fds[tmp->nfds-1].fd = fd;
  tmp->fds[tmp->nfds-1].events = events;
  tmp->data[tmp->nfds-1] = data;

  return 0;
}

int
poller_modify (int inst, int fd, pollerev_t events, pollerdata_t data)
{
  struct poller * tmp;
  int i, index = poller_index (inst);

  if (index < 0)
    return (-1);

  tmp = pollers + index;

  if (fd < 0)
    {
      errno = EBADF;
      return (-1);
    }

  for (i = 0; i < tmp->nfds; i++)
    if (tmp->fds[i].fd == fd)
      break;

  if (i == tmp->nfds)
    {
      errno = ENOENT;
      return (-1);
    }

  tmp->fds[i].events = events;
  tmp->data[i] = data;

  return 0;
}

int
poller_remove (int inst, int fd)
{
  struct poller * tmp;
  int i, index = poller_index (inst);

  if (index < 0)
    return (-1);

  tmp = pollers + index;

  if (fd < 0)
    {
      errno = EBADF;
      return (-1);
    }

  for (i = 0; i < tmp->nfds; i++)
    if (tmp->fds[i].fd == fd)
      break;

  if (i == tmp->nfds)
    {
      errno = ENOENT;
      return (-1);
    }

  -- tmp->nfds;
  if (i < tmp->nfds)
    {
      tmp->fds[i] = tmp->fds[tmp->nfds];
      tmp->data[i] = tmp->data[tmp->nfds];
    }

  tmp->fds = realloc (tmp->fds, sizeof (struct pollfd) * tmp->nfds);
  tmp->data = realloc (tmp->data, sizeof (pollerdata_t) * tmp->nfds);

  return 0;
}

int
poller_poll (int inst, int timeout, pollercb_t callback)
{
  struct poller * tmp;
  int i, res, have, index = poller_index (inst);

  if (index < 0)
    return (-1);

  tmp = pollers + index;

  res = poll (tmp->fds, tmp->nfds, timeout);
  if (res <= 0)
    return res;

  for (i = 0, have = 0; i < tmp->nfds && have < res; i++)
    if (tmp->fds[i].revents)
      {
        callback (tmp->data[i], tmp->fds[i].revents);
        have ++;
      }

  return res;
}

#endif /* not HAVE_EPOLL_WAIT */
